#include <linux/kernel.h>
#include <asm/uaccess.h>
#include <linux/spinlock.h>
#include <linux/proc_fs.h>
#include <linux/fs.h>
#include <linux/timer.h>
#include <linux/module.h>

#include <asm/avalanche/generic/pal.h>
#define ACTION_TEC
#define ACTION_TEC_PUSHBUTTON

#define AR7_RESET_FILE "led_mod/ar7reset"
#define AR7_RESET_GPIO 11
#define RESET_POLL_TIME 1
#ifdef ACTION_TEC
  #ifdef ACTION_TEC_VERIZON
    #define RESET_HOLD_TIME 1
  #else
    #define RESET_HOLD_TIME 10 
  #endif
#else
#define RESET_HOLD_TIME 4
#endif  //ACTION_TEC
#ifdef ACTION_TEC
unsigned int powerred_mod;
#endif //ACTION_TEC

#ifdef ACTION_TEC_PUSHBUTTON
#define AR7_PUSH_FILE   "led_mod/push" 
#define PUSH_POLL_TIME 1
#ifdef ACTION_TEC_QWEST
#define AR7_PUSH_GPIO   14
#define PUSH_HOLD_TIME  8
#else
#define AR7_PUSH_GPIO   2 
#define PUSH_HOLD_TIME  2
#endif
static struct proc_dir_entry *push_file;
//static int count_push=0; 
static int push_count=0;
static int res_state_push = 0;
#endif //ACTION_TEC_PUSHBUTTON

static struct proc_dir_entry *reset_file;
static int res_state = 0;
static int count;
static struct timer_list reset_timer, *pTimer = NULL;
static ssize_t proc_read_reset_fops(struct file *filp,
                                    char *buf,size_t count , loff_t *offp);

struct file_operations reset_fops = {
                                     read: proc_read_reset_fops
                                   };


static ssize_t proc_read_reset_fops(struct file *filp,
                                 char *buf,size_t count , loff_t *offp)
{
  char * pdata = NULL;
  char line[3];
  int len = 0;
  if( *offp != 0 )
  	return 0;

  pdata = buf;
	len = sprintf(line,"%d\n", res_state );
  res_state = 0;
  copy_to_user(buf,line,len );
	*offp = len;
	return len;
}

#ifdef ACTION_TEC_PUSHBUTTON
static struct timer_list push_timer, *pushTimer = NULL;
static ssize_t proc_read_push_fops(struct file *filp,
                                 char *buf,size_t count , loff_t *offp)
{
    char * pdata = NULL;
    char line[3];
    int len = 0;

    if( *offp != 0 )
        return 0;

    pdata = buf;
    len = sprintf(line,"%d\n", res_state_push );
    res_state_push = 0;
    copy_to_user(buf,line,len );
    *offp = len;
    return len;
}

struct file_operations push_fops = {
                                     read: proc_read_push_fops
                                   };

static void push_timer_func(unsigned long data)
{
    //int button_pushed;

    //button_pushed = (PAL_sysGpioInBit(AR7_PUSH_GPIO) == 0);
    //if (button_pushed) count_push++;
    push_count = (PAL_sysGpioInBit(AR7_PUSH_GPIO) == 0) ? push_count + 1: 0;
    //if ((count_push >= PUSH_HOLD_TIME/RESET_POLL_TIME) && !button_pushed )
    //if( (push_count > 0) && (push_count < PUSH_HOLD_TIME/PUSH_POLL_TIME) )
    if(push_count > 0)
    {
        //printk ("kernel: push button pressed ..%d.\n", push_count);
        res_state_push = 1;
    }

    pushTimer->expires = jiffies + HZ*PUSH_POLL_TIME;
    add_timer (pushTimer);
    return;
}
#endif //ACTION_TEC_PUSHBUTTON
static void reset_timer_func(unsigned long data)
{
  count = (PAL_sysGpioInBit(AR7_RESET_GPIO) == 0) ? count + 1: 0;
#ifdef ACTION_TEC  // Apply for all case (normal and VERIZON)
  {
        if( (count > 0) && (count < RESET_HOLD_TIME/RESET_POLL_TIME) ) {
            printk ("Reset key pressed %d\n", count);
            res_state = 1;
        }
        else if(count >= RESET_HOLD_TIME/RESET_POLL_TIME) {
            printk ("Restore to default key pressed %d\n", count);
            res_state = 2;
        }	
		
        pTimer->expires = jiffies + HZ*RESET_POLL_TIME;
        add_timer (pTimer);
        return;
  }
#else // Following former code

#ifdef ACTION_TEC
  {
      static int old_count = 0;
      if ((old_count > 0) && (old_count < RESET_HOLD_TIME/RESET_POLL_TIME) && (count == 0)) {
          printk ("reset key pressed %d\n", old_count);
          machine_restart(NULL);
      }
      old_count = count;
      if ((count==0) && (res_state==1)) {
          res_state = 2;
      }
  }
#endif ACTION_TEC
  if( count >= RESET_HOLD_TIME/RESET_POLL_TIME )
#ifdef ACTION_TEC
  {
      printk ("restore to default key pressed ...\n");
      res_state = 1;
      avalanche_led_action(powerred_mod, 0);
  }
#else ACTION_TEC
    res_state = 1;
#endif ACTION_TEC

        pTimer->expires = jiffies + HZ*RESET_POLL_TIME;
        add_timer (pTimer);
  return;
#endif //ACTION_TEC_VERIZON
}

#ifdef  ACTION_TEC_PUSHBUTTON
void sangam_push_init(void)
{
    printk ("--------init push-----\n");
    /* Create board push proc file */
    push_file = create_proc_entry( AR7_PUSH_FILE, 0777, NULL);
    if( push_file == NULL) {
        printk ("-------failed to push\n");
        remove_proc_entry(AR7_PUSH_FILE,NULL);
        return;
    }
    push_file->owner = THIS_MODULE;
    push_file->proc_fops = &push_fops;

    /* Initialise GPIO 10 for input */
    PAL_sysGpioCtrl(AR7_PUSH_GPIO,GPIO_PIN,GPIO_INPUT_PIN);

  /* Create a timer which fires every seconds */
  pushTimer = &push_timer;
        init_timer( pushTimer );
        pushTimer->function = push_timer_func;
        pushTimer->data = 0;
  /* Start the timer */
  push_timer_func(0);
  return ;
}
#endif

void sangam_reset_init(void)
{
  /* Create board reset proc file */
  proc_mkdir("led_mod",NULL); 
  reset_file = create_proc_entry( AR7_RESET_FILE, 0777, NULL);
  if( reset_file == NULL)
    goto reset_file;
  reset_file->owner = THIS_MODULE;
  reset_file->proc_fops = &reset_fops;

  /* Initialise GPIO 11 for input */
  PAL_sysGpioCtrl(AR7_RESET_GPIO,GPIO_PIN,GPIO_INPUT_PIN);

  /* Create a timer which fires every seconds */
  pTimer = &reset_timer;
	init_timer( pTimer );
	pTimer->function = reset_timer_func;
	pTimer->data = 0;
  /* Start the timer */
  reset_timer_func(0);
#ifdef ACTION_TEC 
  /* register instance 0 of adsl module */   
   powerred_mod = avalanche_led_register("powerred", 0);
#endif //ACTION_TEC

#ifdef  ACTION_TEC_PUSHBUTTON
     sangam_push_init();
#endif
  
  return ;

  reset_file:

    remove_proc_entry("AR7_RESET_FILE",NULL);
	return;
}

#ifdef ACTION_TEC_VERIZON_WRA
#define AR7_STANDBY_GPIO 5
#define AR7_STANDBY_FILE "led_mod/ar7standby"

static struct proc_dir_entry *standby_file;
static struct timer_list standby_timer, *sbTimer = NULL;

static int sbcount;
static int sb_state = 0;
unsigned int standby_mod;
unsigned int pppoegreen_mod;
unsigned int pppoered_mod;
static ssize_t proc_read_standby_fops(struct file *filp,
                                    char *buf,size_t sbcount , loff_t *offp);
                                                                                                                             
struct file_operations standby_fops = {
                                     read: proc_read_standby_fops
                                   };
                                                                                                                             
                                                                                                                             
static ssize_t proc_read_standby_fops(struct file *filp,
                                 char *buf,size_t sbcount , loff_t *offp)
{
  char * pdata = NULL;
  char line[3];
  int len = 0;
  if( *offp != 0 )
        return 0;
  
  pdata = buf;
  len = sprintf(line,"%d\n", sb_state );
//  sb_state = 0;
  copy_to_user(buf,line,len );
   *offp = len;
  return len;
}


static void standby_timer_func(unsigned long data)
{
      sbcount = (PAL_sysGpioInBit(AR7_STANDBY_GPIO) == 0) ? sbcount + 1: 0;
      if (sbcount == 1) {
	  sb_state = 1;
	  if (get_standby() == 0)
	  {
              printk ("enter standby mode\n");
	      avalanche_led_action(standby_mod, 1);
	      set_standby(1);
	      avalanche_led_action_standby(pppoegreen_mod, 1, 1);
	      avalanche_led_action_standby(pppoered_mod, 1, 1);
	  } else
	  {
	      printk ("exit standby mode\n");
	      set_standby(0);
              avalanche_led_action(standby_mod, 0);
	      avalanche_led_action_standby(pppoegreen_mod, 1, 0);
              avalanche_led_action_standby(pppoered_mod, 1, 0);
	  }
      }  
      sbTimer->expires = jiffies + HZ*RESET_POLL_TIME;
      add_timer (sbTimer);
      return;
}

void sangam_standby_init(void)
{
//  proc_mkdir("led_mod",NULL);
  standby_file = create_proc_entry( AR7_STANDBY_FILE, 0777, NULL);
  if( standby_file == NULL)
  {
    goto standby_file;
  } 
  sbcount = 0;
  standby_file->owner = THIS_MODULE;
  standby_file->proc_fops = &standby_fops;
  PAL_sysGpioCtrl(AR7_STANDBY_GPIO,GPIO_PIN,GPIO_INPUT_PIN);
                                                                                                                             
  /* Create a timer which fires every seconds */
  sbTimer = &standby_timer;
        init_timer( sbTimer );
        sbTimer->function = standby_timer_func;
        sbTimer->data = 0;
  /* Start the timer */
  standby_timer_func(0);
  /* register instance 0 of adsl module */
   standby_mod = avalanche_led_register("standby", 0);
   pppoegreen_mod = avalanche_led_register("pppoe", 0);
   pppoered_mod = avalanche_led_register("pppoered", 0);
  return ;

                                                                                                                             
  standby_file:
                                                                                                                             
    remove_proc_entry(AR7_STANDBY_FILE,NULL);
        return;
}

#endif

