
#include <linux/config.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/in.h>
#include <linux/if_vlan.h>

#include <asm/uaccess.h> /* for copy_from_user */

#include <net/arp.h>
#include <net/ip.h>

#ifdef ACTION_TEC_VIRTUAL_ETH
#include <linux/cpmac_veth.h>


struct net_device_stats *cpmac_veth_dev_get_stats(struct net_device *dev)
{
    return &(((struct cpmac_veth_dev_info *)(dev->priv))->dev_stats);
}

int cpmac_veth_dev_hard_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct net_device_stats *stats = cpmac_veth_dev_get_stats(dev);
	struct vlan_ethhdr *veth = (struct vlan_ethhdr *)(skb->data);
#ifdef ACTION_TEC_PIXELATION_SOLUTION
    unsigned char *dest_mac;
    unsigned int mc_data = 0;
#endif /* ACTION_TEC_PIXELATION_SOLUTION */

    int boCopy = 0;
    struct sk_buff* skb_new;

#ifdef ACTION_TEC_PIXELATION_SOLUTION
    dest_mac = skb->data;
    /* Check if this packet is a Multicast Data packet */
    if (
        (dest_mac[0] & 1) &&
        (!dest_mac[1]) &&
        (dest_mac[2] & 0x5e))
    {
        struct iphdr *iph;

	    if (veth->h_vlan_proto == __constant_htons(ETH_P_8021Q))
        {
            /* Packet has VLAN Header */
            if (veth->h_vlan_encapsulated_proto == __constant_htons(ETH_P_IP))
            {
                iph = skb->data+sizeof(struct vlan_ethhdr);
                if (iph->protocol != IPPROTO_IGMP)
                {
                    /*
                    printk("vlan_dev_hard_start_xmit: Multicast data for group=%x to %s\n", iph.daddr, dev->name);
                    printk("hard_start_xmit: VLAN - mc_vlan_ids=%x\n", skb->mc_vlan_ids);
                    */
                    mc_data = 1;
                }
            }
        }
        else
        {
            /* Packet does not have VLAN Header */
            if (veth->h_vlan_proto == __constant_htons(ETH_P_IP))
            {
                iph = skb->data+sizeof(struct ethhdr);
                if (iph->protocol != IPPROTO_IGMP)
                {
                    /*
                    printk("vlan_dev_hard_start_xmit: Multicast data for group=%x to %s\n", iph.daddr, dev->name);
                    printk("hard_start_xmit: NOVLAN - mc_vlan_ids=%x\n", skb->mc_vlan_ids);
                    */
                    mc_data = 1;
                }
            }
        }
    }
#endif /* ACTION_TEC_PIXELATION_SOLUTION */

// Prepare Marvell CPU header
#define MARVEL_HEADER_SIZE 2
#define MIN_ETH_PACKET_LEN 60
  
   /*
    * Here we need to check header insertion and padding, if we have 
    * no place at the beginning of buffer for header, or we have no 
    * sufficient padding space we copy the packet to a fresh SKB 
    */
    if( ((skb->data - skb->head) < MARVEL_HEADER_SIZE) || 
         ((skb->end - skb->head) < MIN_ETH_PACKET_LEN) )
        boCopy = 1;

    if( boCopy ) 
    {
        skb_new = dev_alloc_skb(1536);
        if(skb_new == NULL) 
        { 
            /* silently drop */
            printk(KERN_CRIT "cpmac_veth_dev_hard_start_xmit: Silently drop\n");
            return -EBUSY;
        }
        skb_new->dev = skb->dev;

        memmove( skb_new->data+2, skb->data, skb->len );
        skb_new->len = skb->len+2;
        skb_new->tail = skb_new->data + skb_new->len;

        /* Free the original skb */ 
        dev_kfree_skb(skb);
        skb = skb_new;
    }
    else
    {
        skb->data -= MARVEL_HEADER_SIZE; //boundary problem!
        skb->len += MARVEL_HEADER_SIZE;
    }

    /* 
     * We need to check the length before adding the header's 
     * two bytes, after switch will extract header the packet 
     * (with 4 bytes crc) should be 64 or longer
     */
    if(skb->len < (MIN_ETH_PACKET_LEN+2)) 
    {
        memset(skb->tail, 0, (skb->end-skb->tail));
        /* this is safe because we allocated this skb */
        skb->len = (MIN_ETH_PACKET_LEN+2); 
        skb->tail = skb->data + skb->len;
    }

    *(skb->data) = 0x0; //learn disable and overide vlan //trailer[0];

#ifdef ACTION_TEC_PIXELATION_SOLUTION
    if (mc_data == 1)
    {
        /* 
         * If Multicast data, send it on all the ports in the 
         * Multicast Group 
         */
        *(skb->data+1) = skb->mc_vlan_ids;
    }
    else
#endif /* ACTION_TEC_PIXELATION_SOLUTION */
        *(skb->data+1) = (0x1 << (CPMAC_VETH_DEV_INFO(dev)->port_num));
    
    skb->dev = CPMAC_VETH_DEV_INFO(dev)->real_dev;
#ifdef ACTION_TEC
        spin_lock(&skb->dev->xmit_lock);

if(!strncmp(skb->dev->name,"cp",2)){
	spin_lock(CPMAC_VETH_DEV_INFO(dev)->real_dev->xmit_lock);
	if( skb->dev->hard_start_xmit(skb,skb->dev) != 0)
		printk ("transmit err on cp device\n");
	spin_unlock(CPMAC_VETH_DEV_INFO(dev)->real_dev->xmit_lock);
//	printk(".");
}
else

	if( skb->dev->hard_start_xmit(skb,skb->dev) != 0){
		//I'm   spin_unlock(&skb->dev->xmit_lock);
#ifdef ACTION_TEC_VeriWAVE_DEBUG
		printk ("transmit err\n");
#endif
		//I'm not sure: seems wireless driver will release skb even though error happens, how about ethernet?  kfree_skb(skb);
		//I'm   return -ENETDOWN;
	}

        spin_unlock(&skb->dev->xmit_lock);
#else
	dev_queue_xmit(skb);
#endif
	stats->tx_packets++; 
	stats->tx_bytes += skb->len;
	return 0;
}

int cpmac_veth_dev_change_mtu(struct net_device *dev, int new_mtu)
{
	if (CPMAC_VETH_DEV_INFO(dev)->real_dev->mtu < new_mtu)
		return -ERANGE;

	dev->mtu = new_mtu;

	return new_mtu;
}

int cpmac_veth_dev_open(struct net_device *dev)
{
	if (!(CPMAC_VETH_DEV_INFO(dev)->real_dev->flags & IFF_UP))
		return -ENETDOWN;

	return 0;
}

int cpmac_veth_dev_stop(struct net_device *dev)
{
    if (dev)
	    cpmac_veth_flush_mc_list(dev);

	return 0;
}

int cpmac_veth_dev_init(struct net_device *dev)
{
	return 0;
}

void cpmac_veth_dev_destruct(struct net_device *dev)
{
    if (dev) 
    {
		cpmac_veth_flush_mc_list(dev);

		if (dev->priv) 
        {
			dev_put(CPMAC_VETH_DEV_INFO(dev)->real_dev);

			kfree(dev->priv);

			dev->priv = NULL;
		}
	}
}

int cpmac_veth_dev_set_mac_address(struct net_device *dev, void *addr_struct_p)
{
	int i;
	struct sockaddr *addr = (struct sockaddr *)(addr_struct_p);

	if (netif_running(dev))
		return -EBUSY;

	memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);

	printk("%s: Setting MAC address to ", dev->name);
	for (i = 0; i < 6; i++)
		printk(" %2.2x", dev->dev_addr[i]);
	printk(".\n");

    if (memcmp(CPMAC_VETH_DEV_INFO(dev)->real_dev->dev_addr,
               dev->dev_addr,
               dev->addr_len) != 0) 
    {
        if (!(CPMAC_VETH_DEV_INFO(dev)->real_dev->flags & IFF_PROMISC)) 
        {
			int flgs = CPMAC_VETH_DEV_INFO(dev)->real_dev->flags;

			/* Increment our in-use promiscuity counter */
			dev_set_promiscuity(CPMAC_VETH_DEV_INFO(dev)->real_dev, 1);

			/* Make PROMISC visible to the user. */
			flgs |= IFF_PROMISC;
			printk("VETH (%s):  Setting underlying device (%s) to promiscious mode.\n",
			       dev->name, CPMAC_VETH_DEV_INFO(dev)->real_dev->name);

			dev_change_flags(CPMAC_VETH_DEV_INFO(dev)->real_dev, flgs);
		}
	} 
    else 
    {
		printk("VETH (%s):  Underlying device (%s) has same MAC, not checking promiscious mode.\n",
		       dev->name, CPMAC_VETH_DEV_INFO(dev)->real_dev->name);
	}

	return 0;
}

/********************************************************************** 
 * Multicast support - START
 **********************************************************************/
/* 
 * dmi is a single entry into a dev_mc_list, a single node. mc_list is
 * an entire list, and we'll iterate through it.
 */
int cpmac_veth_should_add_mc(struct dev_mc_list *dmi, 
                             struct dev_mc_list *mc_list)
{
    struct dev_mc_list *idmi;

    for (idmi = mc_list; idmi != NULL; ) 
    {
        if (cpmac_veth_dmi_equals(dmi, idmi)) 
        {
            if (dmi->dmi_users > idmi->dmi_users)
                return 1;
            else
                return 0;
        } 
        else 
        {
            idmi = idmi->next;
        }
    }

    return 1;
}

void cpmac_veth_copy_mc_list(struct dev_mc_list *mc_list, 
                             struct cpmac_veth_dev_info *veth_info)
{
    struct dev_mc_list *dmi, *new_dmi;

    cpmac_veth_destroy_mc_list(veth_info->old_mc_list);
    veth_info->old_mc_list = NULL;

    for (dmi = mc_list; dmi != NULL; dmi = dmi->next) 
    {
        new_dmi = kmalloc(sizeof(*new_dmi), GFP_ATOMIC);
        if (new_dmi == NULL) 
        {
            printk(KERN_ERR "vlan: cannot allocate memory. "
                            "Multicast may not work properly from now.\n");
            return;
        }

        /* Copy whole structure, then make new 'next' pointer */
        *new_dmi = *dmi;
        new_dmi->next = veth_info->old_mc_list;
        veth_info->old_mc_list = new_dmi;
    }
}

void cpmac_veth_dev_set_multicast_list(struct net_device *veth_dev)
{
    struct dev_mc_list *dmi;
    struct net_device *real_dev;
    int inc;

    if (veth_dev && (veth_dev->priv_flags & IFF_VETH)) 
    {
		/* it's a virtual ethernet device */
        real_dev = CPMAC_VETH_DEV_INFO(veth_dev)->real_dev;

		/* compare the current promiscuity to the last promisc we had.. */
		inc = veth_dev->promiscuity - 
                    CPMAC_VETH_DEV_INFO(veth_dev)->old_promiscuity;
		if (inc) 
        {
            printk(KERN_INFO "%s: dev_set_promiscuity(master, %d)\n",
			        veth_dev->name, inc);
            dev_set_promiscuity(real_dev, inc); /* found in dev.c */
            CPMAC_VETH_DEV_INFO(veth_dev)->old_promiscuity = 
                    veth_dev->promiscuity;
		}

        inc = veth_dev->allmulti - CPMAC_VETH_DEV_INFO(veth_dev)->old_allmulti;
        if (inc) 
        {
            printk(KERN_INFO "%s: dev_set_allmulti(master, %d)\n",
			       veth_dev->name, inc);
            dev_set_allmulti(real_dev, inc); /* dev.c */
            CPMAC_VETH_DEV_INFO(veth_dev)->old_allmulti = veth_dev->allmulti;
		}

		/* looking for addresses to add to master's list */
        for (dmi = veth_dev->mc_list; dmi != NULL; dmi = dmi->next) 
        {
            if (cpmac_veth_should_add_mc(dmi, VLAN_DEV_INFO(veth_dev)->old_mc_list)) 
            {
                dev_mc_add(real_dev, dmi->dmi_addr, dmi->dmi_addrlen, 0);
                printk(KERN_INFO "%s: add %.2x:%.2x:%.2x:%.2x:%.2x:%.2x mcast address to master interface\n",
				       veth_dev->name,
				       dmi->dmi_addr[0],
				       dmi->dmi_addr[1],
				       dmi->dmi_addr[2],
				       dmi->dmi_addr[3],
				       dmi->dmi_addr[4],
				       dmi->dmi_addr[5]);
			}
		}

		/* looking for addresses to delete from master's list */
        for (dmi = CPMAC_VETH_DEV_INFO(veth_dev)->old_mc_list; dmi != NULL; dmi = dmi->next) 
        {
            if (cpmac_veth_should_add_mc(dmi, veth_dev->mc_list)) 
            {
				/* if we think we should add it to the new list, then we should really
				 * delete it from the real list on the underlying device.
				 */
                dev_mc_delete(real_dev, dmi->dmi_addr, dmi->dmi_addrlen, 0);
                printk(KERN_INFO "%s: del %.2x:%.2x:%.2x:%.2x:%.2x:%.2x mcast address from master interface\n",
				       veth_dev->name,
				       dmi->dmi_addr[0],
				       dmi->dmi_addr[1],
				       dmi->dmi_addr[2],
				       dmi->dmi_addr[3],
				       dmi->dmi_addr[4],
				       dmi->dmi_addr[5]);
			}
		}

		/* save multicast list */
        cpmac_veth_copy_mc_list(veth_dev->mc_list, 
                                CPMAC_VETH_DEV_INFO(veth_dev));
	}
}

void cpmac_veth_flush_mc_list(struct net_device *dev)
{
    struct dev_mc_list *dmi = dev->mc_list;

    while (dmi) 
    {
        dev_mc_delete(dev, dmi->dmi_addr, dmi->dmi_addrlen, 0);
        printk(KERN_INFO "%s: del %.2x:%.2x:%.2x:%.2x:%.2x:%.2x mcast address from vlan interface\n",
		       dev->name,
		       dmi->dmi_addr[0],
		       dmi->dmi_addr[1],
		       dmi->dmi_addr[2],
		       dmi->dmi_addr[3],
		       dmi->dmi_addr[4],
		       dmi->dmi_addr[5]);
        dmi = dev->mc_list;
    }

	/* dev->mc_list is NULL by the time we get here. */
    cpmac_veth_destroy_mc_list(CPMAC_VETH_DEV_INFO(dev)->old_mc_list);
    CPMAC_VETH_DEV_INFO(dev)->old_mc_list = NULL;
}
/********************************************************************** 
 * Multicast support - END
 **********************************************************************/
#endif /* ACTION_TEC_VIRTUAL_ETH */
