#ifndef _CPMAC_VETH_H_
#define _CPMAC_VETH_H_

#ifdef __KERNEL__

#ifdef ACTION_TEC_VIRTUAL_ETH

#define CPMAC_VETH_DEBUG

struct cpmac_veth_dev_info 
{
	unsigned int port_num;      /*  Physical port number */
	unsigned short flags;       

    /*
     * old multi-cast list for the interface..
     * we save this so we can tell what changes were
     * made, in order to feed the right changes down
     * to the real hardware...
     */
	struct dev_mc_list *old_mc_list;  
	int old_allmulti;               /* similar to above. */
	int old_promiscuity;            /* similar to above. */

	struct net_device *real_dev;    /* the underlying device/interface */

    /* Device stats */
	struct net_device_stats dev_stats; 
};

#define CPMAC_VETH_DEV_INFO(x) ((struct cpmac_veth_dev_info *)(x->priv))

#define CPMAC_VETH_DEV_NUM      4
extern struct net_device *cpmac_veth_devices[CPMAC_VETH_DEV_NUM];

static inline int cpmac_veth_dmi_equals(struct dev_mc_list *dmi1,
                                        struct dev_mc_list *dmi2)
{
    return ((dmi1->dmi_addrlen == dmi2->dmi_addrlen) && 
            (memcmp(dmi1->dmi_addr, dmi2->dmi_addr, dmi1->dmi_addrlen) == 0));
}

static inline void cpmac_veth_destroy_mc_list(struct dev_mc_list *mc_list)
{
    struct dev_mc_list *dmi = mc_list;
    struct dev_mc_list *next;

    while(dmi) 
    {
        next = dmi->next;
        kfree(dmi);
        dmi = next;
    }
}

/* From cpmac_veth.c */
struct net_device *register_cpmac_veth_device(const char *eth_if_name, 
                                              unsigned int port_num);
int unregister_cpmac_veth_device(const char *veth_if_name);

/* From cpmac_veth_dev.c */
struct net_device_stats *cpmac_veth_dev_get_stats(struct net_device *dev);

int cpmac_veth_dev_hard_start_xmit(struct sk_buff *skb, struct net_device *dev);

int cpmac_veth_dev_change_mtu(struct net_device *dev, int new_mtu);
int cpmac_veth_dev_open(struct net_device *dev);
int cpmac_veth_dev_stop(struct net_device *dev);
int cpmac_veth_dev_init(struct net_device *dev);
void cpmac_veth_dev_destruct(struct net_device *dev);

int cpmac_veth_dev_set_mac_address(struct net_device *dev, void *addr_struct_p);

int cpmac_veth_should_add_mc(struct dev_mc_list *dmi, 
                             struct dev_mc_list *mc_list);
void cpmac_veth_copy_mc_list(struct dev_mc_list *mc_list, 
                             struct cpmac_veth_dev_info *veth_info);
void cpmac_veth_dev_set_multicast_list(struct net_device *veth_dev);
void cpmac_veth_flush_mc_list(struct net_device *dev);
#endif /* ACTION_TEC_VIRTUAL_ETH */

#endif /* __KERNEL__ */

#endif /* _CPMAC_VETH_H_ */
