/* vi: set sw=4 ts=4: */
/*
 * Mini nslookup implementation for busybox
 *
 * Copyright (C) 1999,2000 by Lineo, inc. and John Beppu
 * Copyright (C) 1999,2000,2001 by John Beppu <beppu@codepoet.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <resolv.h>
#include <arpa/inet.h>
#include "busybox.h"

/*
 |  I'm only implementing non-interactive mode;
 |  I totally forgot nslookup even had an interactive mode.
 |
 |  [ TODO ]
 |  + find out how to use non-default name servers
 */


 /*
 * Copyright (c) 1985, 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *	  must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.	IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
/*static char *sccsid = "from: @(#)res_search.c	6.45 (Berkeley) 2/24/91";*/
static char *rcsid = "$Id: nslookup.c,v 1.1.1.1.58.8 2010/08/12 01:08:17 denny Exp $";
#endif /* LIBC_SCCS and not lint */

#include <pthread.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <resolv.h>
#include <netdb.h>
#include "res_internal.h"

#define MAXPACKET 65468
#define __NETDB_MAXADDRS 35
#define __NETDB_MAXALIASES 35
#define pthread_ipaddr_type int
#define HOST_BUFSIZE 4096
#define ALIGN(p, t) ((char *)(((((long)(p) - 1) / sizeof(t)) + 1) * sizeof(t)))
#define SP(p, t, n) (ALIGN(p, t) + (n) * sizeof(t))

struct __res_state _res = {
        RES_TIMEOUT,                    /* retransmition time interval */
        4,                              /* number of times to retransmit */
        RES_DEFAULT,                            /* options flags */
        1,                              /* number of name servers */
};

/* Performs global initialization. */
struct res_data *globledata = NULL;
static struct __res_state start;
char dns_txt[1024] = {0};
struct res_data *_res_init()
{
        struct res_data *data;
        //printf("--_res_init\n");

        /* Initialize thread-specific data for this thread if it hasn't
         * been done already. */
        if (globledata)
        return globledata;
        if (globledata == NULL)
                globledata = (struct res_data *) malloc(sizeof(struct res_data));
        if (globledata == NULL)
                return NULL;

        globledata->buf = NULL;
        globledata->state = start;
        globledata->errval = NO_RECOVERY;
        globledata->sock = -1;

        //printf("--return globledata\n");
        return globledata;
}

#if 0
int res_init()
{
	if (_res_init()) {
		return 0;
	} else {
		/* Due to clever tricks in _res_init(), a check for h_errno will
		 * return NO_RECOVERY even if the next try at initialization
		 * succeeds, so it's okay that we can't set an error value here. */
		return -1;
	}
}

#endif

static char *search_aliases(const char *name, char *buf, int bufsize);

int res_search(const char *name, int class, int type, unsigned char *answer,
			   int anslen)
{
	struct res_data *data;
	const char *p;
	int num_dots, len, result, no_data = 0, error;
	char buf[2 * MAXDNAME + 2], *domain, **dptr, *alias;

    //printf("--res_search name=[%s]\n", name);

	data = _res_init();
	if (!data)
		return -1;

	/* Count the dots in name, and get a pointer to the end of name. */
	num_dots = 0;
	for (p = name; *p; p++) {
		if (*p == '.')
			num_dots++;
	}
	len = p - name;

    //printf("--len=%d\n", len);

	/* If there aren't any dots, check to see if name is an alias for
	 * another host.  If so, try the resolved alias as a fully-qualified
	 * name. */
	alias = search_aliases(name, buf, sizeof(buf));
    //printf("--alias=[%s]\n", alias);
	if (alias != NULL)
		return res_query(alias, class, type, answer, anslen);

    //printf("--there's a trailing dot\n");

	/* If there's a trailing dot, try to strip it off and query the name. */
	if (len > 0 && p[-1] == '.') {
		if (len > sizeof(buf)) {
			/* It's too long; just query the original name. */
			return res_query(name, class, type, answer, anslen);
		} else {
			/* Copy the name without the trailing dot and query. */
			memcpy(buf, name, len - 1);
			buf[len] = 0;
			return res_query(buf, class, type, answer, anslen);
		}
	}

    //printf("--data->state.options & RES_DNSRCH=%d\n", data->state.options & RES_DNSRCH);
	if (data->state.options & RES_DNSRCH) {
		/* If RES_DNSRCH is set, query all the domains until we get a
		 * definitive answer. */
        //printf("--RES_DNSRCH is set\n");
		for (dptr = data->state.dnsrch; *dptr; dptr++) {
			domain = *dptr;
			sprintf(buf, "%.*s.%.*s", MAXDNAME, name, MAXDNAME, domain);
            //printf("--buf=[%s]\n", buf);
			result = res_query(buf, class, type, answer, anslen);
			if (result > 0)
				return result;
			if (data->errval == NO_DATA)
				no_data = 1;
			else if (data->errval != HOST_NOT_FOUND)
				break;
		}
	} else if (num_dots == 0 && data->state.options & RES_DEFNAMES) {
		/* If RES_DEFNAMES is set and there is no dot, query the default
		 * domain. */
        //printf("-- RES_DEFNAMES is set\n");
		domain = data->state.defdname;
		sprintf(buf, "%.*s.%.%s", MAXDNAME, name, MAXDNAME, domain);
        //printf ("-RES_DEFNAMES buf=[%s]\n", buf);
		result = res_query(buf, class, type, answer, anslen);
        //printf("--result=%d\n", result);
		if (result > 0)
			return result;
		if (data->errval == NO_DATA)
			no_data = 1;
	}

    //printf("--num_dots=%d\n", num_dots);

	/* If all the domain queries failed, try the name as fully-qualified.
	 * Only do this if there is at least one dot in the name. */
	if (num_dots > 0) {
		result = res_query(name, class, type, answer, anslen);
        //printf("--result=%d\n", result);
		if (result > 0)
			return result;
	}

	if (no_data)
		data->errval = NO_DATA;

	return -1;
}

static char *search_aliases(const char *name, char *buf, int bufsize)
{
	FILE *fp;
	char *filename, *p;
	int len;

    //printf("--search_aliases name=%s\n", name);
	filename = getenv("HOSTALIASES");
	if (filename == NULL)
		return NULL;

	fp = fopen(filename, "r");
	if (fp == NULL)
		return NULL;

	len = strlen(name);
	while (fgets(buf, bufsize, fp)) {

		/* Get the first word from the buffer. */
		p = buf;
		while (*p && !isspace(*p))
			p++;
		if (!*p)
			break;

		/* Null-terminate the first word and compare it with the name. */
		*p = 0;
		if (strcasecmp(buf, name) != 0)
			continue;

		p++;
		while (isspace(*p))
			p++;
		fclose(fp);
		return (*p) ? p : NULL;
	}

	fclose(fp);
	return NULL;
}

u_short
_getshort(msgp)
        register const u_char *msgp;
{
        register u_short u;

        GETSHORT(u, msgp);
        return (u);
}

static int printable(int ch) 
{
        return (ch > 0x20 && ch < 0x7f);
}

static int __set_errno (int ch)
{
  printf ("__set_errno %d\n", ch);
  return 0;
}

/*
 * special(ch)
 *      Thinking in noninternationalized USASCII (per the DNS spec),
 *      is this characted special ("in need of quoting") ?
 * return:
 *      boolean.
 */
static int special(int ch) 
{
        switch (ch) {
        case 0x22: /* '"' */
        case 0x2E: /* '.' */
        case 0x3B: /* ';' */
        case 0x5C: /* '\\' */
        /* Special modifiers in zone files. */
        case 0x40: /* '@' */
        case 0x24: /* '$' */
                return (1);
        default:
                return (0);
        }
}

/*
 * Skip over a compressed domain name. Return the size or -1.
 */
__dn_skipname(comp_dn, eom)
        const u_char *comp_dn, *eom;
{
        register u_char *cp;
        register int n;

        //printf("--__dn_skipname\n");
        cp = (u_char *)comp_dn;
        while (cp < eom && (n = *cp++)) {
                /*
                 * check for indirection
                 */
                //printf("--type n=%x INDIR_MASK=%x n & INDIR_MASK=%d\n", n, INDIR_MASK, n & INDIR_MASK);
                switch (n & INDIR_MASK) {
                  case 0:                               /* normal case, n == len */
                        cp += n;
                        continue;
                  case INDIR_MASK:              /* indirection */
                        cp++;
                        break;
                  default:                              /* illegal type */
                        //printf("--illegal type n=%x INDIR_MASK=%x\n", n, INDIR_MASK);
                        return (-1);
                }
                break;
        }
        if (cp > eom)
        {
                //printf("--__dn_skipname return -1\n");
                return -1;
        }
        //printf("--__dn_skipname return %d\n", cp - comp_dn);
        return (cp - comp_dn);
}

/*
 * ns_name_ntop(src, dst, dstsiz)
 *      Convert an encoded domain name to printable ascii as per RFC1035.
 * return:
 *      Number of bytes written to buffer, or -1 (with errno set)
 * notes:
 *      The root is returned as "."
 *      All other domains are returned in non absolute form
 */
int __ns_name_ntop(const u_char *src, char *dst, size_t dstsiz) {
	const u_char *cp;
	char *dn, *eom;
	u_char c;
	u_int n;
	const char digits[] = "0123456789";

	cp = src;
	dn = dst;
	eom = dst + dstsiz;

	while ((n = *cp++) != 0) {
		if ((n & NS_CMPRSFLGS) != 0) {
			/* Some kind of compression pointer. */
			__set_errno (EMSGSIZE);
			return (-1);
		}
		if (dn != dst) {
			if (dn >= eom) {
				__set_errno (EMSGSIZE);
				return (-1);
			}
			*dn++ = '.';
		}
		if (dn + n >= eom) {
			__set_errno (EMSGSIZE);
			return (-1);
		}
		for ((void)NULL; n > 0; n--) {
			c = *cp++;
			if (special(c)) {
				if (dn + 1 >= eom) {
					__set_errno (EMSGSIZE);
					return (-1);
				}
				*dn++ = '\\';
				*dn++ = (char)c;
			} else if (!printable(c)) {
				if (dn + 3 >= eom) {
					__set_errno (EMSGSIZE);
					return (-1);
				}
				*dn++ = '\\';
				*dn++ = digits[c / 100];
				*dn++ = digits[(c % 100) / 10];
				*dn++ = digits[c % 10];
			} else {
				if (dn >= eom) {
					__set_errno (EMSGSIZE);
					return (-1);
				}
				*dn++ = (char)c;
			}
		}
	}
	if (dn == dst) {
		if (dn >= eom) {
			__set_errno (EMSGSIZE);
			return (-1);
		}
		*dn++ = '.';
	}
	if (dn >= eom) {
		__set_errno (EMSGSIZE);
		return (-1);
	}
	*dn++ = '\0';
        return (dn - dst);
}


/*
 * ns_name_unpack(msg, eom, src, dst, dstsiz)
 *      Unpack a domain name from a message, source may be compressed.
 * return:
 *      -1 if it fails, or consumed octets if it succeeds.
 */
int __ns_name_unpack(const u_char *msg, const u_char *eom, const u_char *src,
               u_char *dst, size_t dstsiz)
{
	const u_char *srcp, *dstlim;
	u_char *dstp;
	int n, len, checked;

	len = -1;
	checked = 0;
	dstp = dst;
	srcp = src;
	dstlim = dst + dstsiz;
	if (srcp < msg || srcp >= eom) {
		__set_errno (EMSGSIZE);
		return (-1);
	}
	/* Fetch next label in domain name. */
	while ((n = *srcp++) != 0) {
		/* Check for indirection. */
		switch (n & NS_CMPRSFLGS) {
		case 0:
			/* Limit checks. */
			if (dstp + n + 1 >= dstlim || srcp + n >= eom) {
				__set_errno (EMSGSIZE);
				return (-1);
			}
			checked += n + 1;
			*dstp++ = n;
			memcpy(dstp, srcp, n);
			dstp += n;
			srcp += n;
			break;

		case NS_CMPRSFLGS:
			if (srcp >= eom) {
				__set_errno (EMSGSIZE);
				return (-1);
			}
			if (len < 0)
				len = srcp - src + 1;
			srcp = msg + (((n & 0x3f) << 8) | (*srcp & 0xff));
			if (srcp < msg || srcp >= eom) {  /* Out of range. */
				__set_errno (EMSGSIZE);
				return (-1);
			}
			checked += 2;
			/*
			 * Check for loops in the compressed name;
			 * if we've looked at the whole message,
			 * there must be a loop.
			 */
			if (checked >= eom - msg) {
				__set_errno (EMSGSIZE);
				return (-1);
			}
			break;

		default:
			__set_errno (EMSGSIZE);
			return (-1);                    /* flag error */
		}
	}
	*dstp = '\0';
	if (len < 0)
		len = srcp - src;
	return (len);
}


struct __res_state *_res_status()
{
        struct res_data *data;

        data = _res_init();
        return (data) ? &data->state : NULL;
}

static int qcomp(const void *arg1, const void *arg2)
{
        const struct in_addr **a1 = (const struct in_addr **) arg1;
        const struct in_addr **a2 = (const struct in_addr **) arg2;
        struct __res_state *state = _res_status();

        int pos1, pos2;

        for (pos1 = 0; pos1 < state->nsort; pos1++) {
                if (state->sort_list[pos1].addr.s_addr ==
                        ((*a1)->s_addr & state->sort_list[pos1].mask))
                        break;
        }
        for (pos2 = 0; pos2 < state->nsort; pos2++) {
                if (state->sort_list[pos2].addr.s_addr ==
                        ((*a2)->s_addr & state->sort_list[pos2].mask))
                        break;
        }
        return pos1 - pos2;
}

int __ns_name_uncompress(const u_char *msg, const u_char *eom, 
		const u_char *src, char *dst, size_t dstsiz)
{
	u_char tmp[NS_MAXCDNAME];
	int n;

	if ((n = ns_name_unpack(msg, eom, src, tmp, sizeof tmp)) == -1)
		return (-1);
	if (ns_name_ntop(tmp, dst, dstsiz) == -1)
		return (-1);
	return (n);
}


int __dn_expand(const u_char *msg, const u_char *eom, const u_char *src,
          char *dst, int dstsiz)
{
	int n = ns_name_uncompress(msg, eom, src, dst, (size_t)dstsiz);

    //printf("-dst=[%s]\n", dst);
	if (n > 0 && dst[0] == '.')
		dst[0] = '\0';

    //printf("-__dn_expand n=%d\n", n);
	return (n);
}


struct hostent *_res_parse_answer(querybuf *answer, int anslen, int iquery,
                                                                  struct hostent *result, char *buf,
                                                                  int bufsize, int *errval)
{
        struct res_data *data = _res_init();
        register HEADER *hp;
        register u_char *cp;
        register int n;
        u_char *eom;
        char *aliases[__NETDB_MAXALIASES], *addrs[__NETDB_MAXADDRS];
        char *bp = buf, **ap = aliases, **hap = addrs;
        int type, class, ancount, qdcount, getclass = C_ANY, iquery_done = 0;
        int i, j;
        FILE *fp;

        //printf("--_res_parse_answer iquery=%d anslen=%d header=%d\n", iquery, anslen, sizeof(HEADER));

        eom = answer->buf + anslen;

        fp = fopen ("/var/tmp/dns_txt", "w");
        if (fp)
        {
            for (i=1; i<anslen; i++)
            {
                //printf("%02x=%c ", answer->buf[i], answer->buf[i]);
                fprintf(fp, "%c", answer->buf[i]);
                j++;
                if ( (j % 8) == 0)
                {
                    j = 0;
                    //printf("\n");
                }
            }
            fclose (fp);
        }

        /*
         * find first satisfactory answer
         */
        hp = &answer->hdr;
        ancount = ntohs(hp->ancount);
        qdcount = ntohs(hp->qdcount);
        bp = buf;
        cp = answer->buf + sizeof(HEADER);

        //printf("--ancount=%d qdcount=%d\n", ancount, qdcount);

        /* Read in the hostname if this is an address lookup. */
        if (qdcount) {
                if (iquery) {
                        if ((n = dn_expand((u_char *) answer->buf,
                                                           (u_char *) eom, (u_char *) cp, (u_char *) bp,
                                                           bufsize - (bp - buf))) < 0) {
                                *errval = NO_RECOVERY;
                                //printf("--*errval = NO_RECOVER\n");
                                return ((struct hostent *) NULL);
                        }
                        cp += n + QFIXEDSZ;
                        result->h_name = bp;
                        bp += strlen(bp) + 1;
                } else {
                        cp += __dn_skipname(cp, eom) + QFIXEDSZ;
                        //printf("--cp+=%d\n", QFIXEDSZ);
                }
                while (--qdcount > 0)
                        cp += __dn_skipname(cp, eom) + QFIXEDSZ;
        } else if (iquery) {
                *errval = (hp->aa) ? HOST_NOT_FOUND : TRY_AGAIN;
                //printf("-HOST_NOT_FOUND *errval=%d\n", *errval);
                return ((struct hostent *) NULL);
        }

        //printf("--after ancount=%d qdcount=%d\n", ancount, qdcount);
        //printf("-iquery=%d ancount=%d\n", iquery, ancount);
        //printf("--ancount=%d cp<eom=%d \n", ancount, cp < eom);

        //cp = answer->buf + sizeof(HEADER);  //debug

        /* Read in the answers. */
        *ap = NULL;
        *hap = NULL;
        while (--ancount >= 0 && cp < eom) {
                //printf("--while ancount=%d cp<eom=%d \n", ancount, cp < eom);
                if ((n = dn_expand((u_char *) answer->buf, (u_char *) eom,
                                                   (u_char *) cp, (u_char *) bp,
                                                   bufsize - (bp - buf))) < 0)
                {
                        //printf("--n=%d\n", n);
                        break;
                }
                cp += n;
                type = _getshort(cp);
                cp += sizeof(u_short);
                class = _getshort(cp);
                cp += sizeof(u_short) + sizeof(pthread_ipaddr_type);
                n = _getshort(cp);
                cp += sizeof(u_short);

                //printf("iquery=%d type=%x\n", iquery, type);
                if (type == T_CNAME) {
                        cp += n;
                        if (ap >= aliases + __NETDB_MAXALIASES - 1)
                                continue;
                        *ap++ = bp;
                        bp += strlen(bp) + 1;
                        continue;
                }
                if (iquery && type == T_PTR) {
                        if ((n = dn_expand((u_char *) answer->buf, (u_char *) eom,
                                                           (u_char *) cp, (u_char *) bp,
                                                           bufsize - (bp - buf))) < 0)
                                break;
                        cp += n;
                        result->h_name = bp;
                        bp += strlen(bp) + 1;
                        iquery_done = 1;
                        break;
                }

                if (iquery || type == 0x0010)   {
                        //printf("expected answer type %d, size %d\n", type, n);
                        FILE *fp = NULL;
                        strncpy(dns_txt, cp, n);
                        dns_txt[n] = '\0';
                        //printf("dns txt %s\n", dns_txt);
                        fp = fopen ("/var/tmp/dns_txt", "w");
                        if (fp)
                        {
                            fprintf(fp, "%s", dns_txt);
                            fclose (fp);
                        }
                        cp += n;
                        //iquery_done = 1;
                        //continue;
                }

                if (hap > addrs) {
                        if (n != result->h_length) {
                                cp += n;
                                continue;
                        }
                        if (class != getclass) {
                                cp += n;
                                continue;
                        }
                } else {
                        result->h_length = n;
                        getclass = class;
                        result->h_addrtype = (class == C_IN) ? AF_INET : AF_UNSPEC;
                        if (!iquery) {
                                result->h_name = bp;
                                bp += strlen(bp) + 1;
                        }
                }

                bp = ALIGN(bp, char *);
                //bp = ALIGN(bp, pthread_ipaddr_type);
                if (bp + n >= buf + bufsize) {
                        errno = ERANGE;
                        return NULL;
                }
                memcpy(bp, cp, n);
                cp += n;
                if (hap >= addrs + __NETDB_MAXADDRS - 1)
                        continue;
                *hap++ = bp;
                bp += n;
                cp += n;
        }

        if (hap > addrs || iquery_done) {
                *ap++ = NULL;
                *hap++ = NULL;
                if (data->state.nsort)
                        qsort(addrs, hap - addrs, sizeof(struct in_addr), qcomp);
                if (SP(bp, char *, (hap - addrs) + (ap - aliases)) > buf + bufsize) {
                        errno = ERANGE;
                        return NULL;
                }
                result->h_addr_list = (char **) ALIGN(bp, char *);
                memcpy(result->h_addr_list, addrs, (hap - addrs) * sizeof(char *));
                result->h_aliases = result->h_addr_list + (hap - addrs);
                memcpy(result->h_aliases, aliases, (ap - aliases) * sizeof(char *));
                return result;
        } else {
                *errval = TRY_AGAIN;
                return NULL;
        }
}

static struct hostent *fake_hostent(const char *hostname, struct in_addr addr,
                                                                        struct hostent *result, char *buf,
                                                                        int bufsize, int *errval);
static struct hostent *file_find_name(const char *name, struct hostent *result,
                                                                          char *buf, int bufsize, int *errval);
#if 0
static struct hostent *file_find_name(const char *name, struct hostent *result,
									  char *buf, int bufsize, int *errval)
{
	char **alias;
	FILE *fp = NULL;

	pthread_mutex_lock(&host_iterate_lock);
	sethostent(0);

/*
    extern int gethostent_r (struct hostent *__restrict __result_buf,
                         char *__restrict __buf, size_t __buflen,
                         struct hostent **__restrict __result,
                         int *__restrict __h_errnop) __THROW;
*/


	//while ((result = gethostent_r(result, buf, bufsize, errval)) != NULL) 
    while (gethostent_r(result, buf, bufsize, &result, errval) == 0)
    {
		/* Check the entry's name and aliases against the given name. */
		if (strcasecmp(result->h_name, name) == 0)
			break;
		for (alias = result->h_aliases; *alias; alias++) {
			if (strcasecmp(*alias, name) == 0)
				break;
		}
	}
	pthread_mutex_unlock(&host_iterate_lock);
	if (!result && errno != ERANGE)
		*errval = HOST_NOT_FOUND;
	return result;
}
#endif

static struct hostent *gethostbyname_txt_r(const char *hostname, struct hostent *result,
                                                                char *buf, int bufsize, int *errval)
{
        struct in_addr addr;
        querybuf qbuf;
        const char *p;
        int n;

        //printf("--gethostbyname_txt_r %s\n", hostname);

        /* Default failure condition is not a range error and not recoverable. */
        errno = 0;
        *errval = NO_RECOVERY;

        /* Check for all-numeric hostname with no trailing dot. */
        if (isdigit(hostname[0])) {
                p = hostname;
                while (*p && (isdigit(*p) || *p == '.'))
                        p++;
                if (!*p && p[-1] != '.') {
                        /* Looks like an IP address; convert it. */
                        if (inet_aton(hostname, &addr) == -1) {
                                *errval = HOST_NOT_FOUND;
                                return NULL;
                        }
                        return fake_hostent(hostname, addr, result, buf, bufsize, errval);
                }
        }

        /* Do the search. */
        //n = res_search(hostname, C_IN, T_A, qbuf.buf, sizeof(qbuf));
        n = res_search(hostname, C_IN, 0x0010, qbuf.buf, sizeof(qbuf));
        //printf("--n=%d errno=%d ECONNREFUSED=%d\n", n, errno, ECONNREFUSED);
        if (n >= 0)
                return _res_parse_answer(&qbuf, n, 0, result, buf, bufsize, errval);
#if 0
        else if (errno == ECONNREFUSED)
                return file_find_name(hostname, result, buf, bufsize, errval);
#endif
        else
                return NULL;
}

static struct hostent *fake_hostent(const char *hostname, struct in_addr addr,
                                                                        struct hostent *result, char *buf,
                                                                        int bufsize, int *errval)
{
        int len = strlen(hostname);
        char *name, *addr_ptr;

        //printf("-fake_hostent %d\n", hostname);

        if (SP(SP(SP(buf, char, len + 1), addr, 1), char *, 3) > buf + bufsize) {
                errno = ERANGE;
                return NULL;
        }

        /* Copy faked name and address into buffer. */
        strcpy(buf, hostname);
        name = buf;
        buf = ALIGN(buf + len + 1, addr);
        *((struct in_addr *) buf) = addr;
        addr_ptr = buf;
        buf = ALIGN(buf + sizeof(addr), char *);
        ((char **) buf)[0] = addr_ptr;
        ((char **) buf)[1] = NULL;
        ((char **) buf)[2] = NULL;

        result->h_name = name;
        result->h_aliases = ((char **) buf) + 2;
        result->h_addrtype = AF_INET;
        result->h_length = sizeof(addr);
        result->h_addr_list = (char **) buf;

        return result;
}



static struct hostent *gethostbyname_txt(const char *hostname)
{
        struct res_data *data = _res_init();


        //printf("--gethostbyname_txt %s\n", hostname);
        if (!data)
                return NULL;
        if (!data->buf) {
                data->buf = malloc(sizeof(struct hostent) + HOST_BUFSIZE);
                if (!data->buf) {
                        errno = 0;
                        data->errval = NO_RECOVERY;
                        return NULL;
                }
        }
        return gethostbyname_txt_r(hostname, (struct hostent *) data->buf,
                                                   data->buf + sizeof(struct hostent), HOST_BUFSIZE,
                                                   &data->errval);
}

 


/* only works for IPv4 */
static int addr_fprint(char *addr)
{
	u_int8_t split[4];
	u_int32_t ip;
	u_int32_t *x = (u_int32_t *) addr;

	ip = ntohl(*x);
	split[0] = (ip & 0xff000000) >> 24;
	split[1] = (ip & 0x00ff0000) >> 16;
	split[2] = (ip & 0x0000ff00) >> 8;
	split[3] = (ip & 0x000000ff);
	printf("%d.%d.%d.%d", split[0], split[1], split[2], split[3]);
	return 0;
}

/* takes the NULL-terminated array h_addr_list, and
 * prints its contents appropriately
 */
static int addr_list_fprint(char **h_addr_list)
{
	int i, j;
	char *addr_string = (h_addr_list[1])
		? "Addresses: " : "Address:   ";

	printf("%s ", addr_string);
	for (i = 0, j = 0; h_addr_list[i]; i++, j++) {
		addr_fprint(h_addr_list[i]);

		/* real nslookup does this */
		if (j == 4) {
			if (h_addr_list[i + 1]) {
				printf("\n          ");
			}
			j = 0;
		} else {
			if (h_addr_list[i + 1]) {
				printf(", ");
			}
		}

	}
	printf("\n");
	return 0;
}

/* print the results as nslookup would */
static struct hostent *hostent_fprint(struct hostent *host, const char *server_host)
{
	if (host) {
		printf("%s     %s\n", server_host, host->h_name);
		addr_list_fprint(host->h_addr_list);
	} else {
		printf("*** Unknown host\n");
	}
	return host;
}

/* changes a c-string matching the perl regex \d+\.\d+\.\d+\.\d+
 * into a u_int32_t
 */
static u_int32_t str_to_addr(const char *addr)
{
	u_int32_t split[4];
	u_int32_t ip;

	sscanf(addr, "%d.%d.%d.%d",
		   &split[0], &split[1], &split[2], &split[3]);

	/* assuming sscanf worked */
	ip = (split[0] << 24) |
		(split[1] << 16) | (split[2] << 8) | (split[3]);

	return htonl(ip);
}

/* gethostbyaddr wrapper */
static struct hostent *gethostbyaddr_wrapper(const char *address)
{
	struct in_addr addr;

	addr.s_addr = str_to_addr(address);
	return gethostbyaddr((char *) &addr, 4, AF_INET);	/* IPv4 only for now */
}

/* lookup the default nameserver and display it */
static inline void server_print(void)
{
	struct sockaddr_in def = _res.nsaddr_list[0];
	char *ip = inet_ntoa(def.sin_addr);

    //printf("--server_print ip=[%s]\n", ip);

	hostent_fprint(gethostbyaddr_wrapper(ip), "Server:");
	printf("\n");
}

/* alter the global _res nameserver structure to use
   an explicit dns server instead of what is in /etc/resolv.h */
static inline void set_default_dns(char *server)
{
        struct in_addr server_in_addr;

        if(inet_aton(server,&server_in_addr))
        {
                _res.nscount = 1;
                _res.nsaddr_list[0].sin_addr = server_in_addr;
        }
}

/* naive function to check whether char *s is an ip address */
static int is_ip_address(const char *s)
{
	while (*s) {
		if ((isdigit(*s)) || (*s == '.')) {
			s++;
			continue;
		}
		return 0;
	}
	return 1;
}

/* ________________________________________________________________________ */
int nslookup_main(int argc, char **argv)
{
	struct hostent *host;
/*
	if (argc < 2 || *argv[1]=='-') {
		show_usage();
	}
*/

        if (argc < 2 || *argv[1]=='-' || argc > 3)
                show_usage();
        else if(argc == 3)
                set_default_dns(argv[2]);

	res_init();
	server_print();
	if (is_ip_address(argv[1])) {
		host = gethostbyaddr_wrapper(argv[1]);
	} else {
        host = gethostbyname_txt(argv[1]);
		//host = xgethostbyname(argv[1]);
	}
	hostent_fprint(host, "Name:  ");
	return EXIT_SUCCESS;
}

/* $Id: nslookup.c,v 1.1.1.1.58.8 2010/08/12 01:08:17 denny Exp $ */
