/*
 * This contains _two_ implementations of ps for Linux.  One uses the
 * traditional /proc virtual filesystem, and the other use the devps kernel
 * driver (written by Erik Andersen to avoid using /proc thereby saving 100k+).
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <stdarg.h>
#include <signal.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#ifdef STANDALONE
#define kprintf printf
#else 
#include <fcntl.h>
#define LED_KPRINTF 8
static int kprintf (const char *fmt, ...)
{
    static char tmp_text[51];
	va_list ap;
    int fp;

    /* Open /dev/led file for LED Operation */
    fp = open("/dev/led", O_RDWR);
    if(!fp) {
        return -1;
    }
    
    memset (tmp_text, 0, sizeof (tmp_text));
    va_start(ap, fmt);
    vsnprintf(tmp_text, sizeof (tmp_text)-1, fmt, ap);
    va_end(ap);	
    ioctl(fp, LED_KPRINTF, &tmp_text);
    close(fp);

    return 0;
}
#endif //STANDALONE


/*
 * if ipaddr is in /proc/net/arp, return 1 and copy its mac to buf mac
    or return 0 and set mac buf to space 
*/
static int read_mac(const char* ipaddr, char* mac)
{
    int res = 0;
    char buf[4096];
    char *pos, *ptr;
    int fd;
    
    sprintf(mac, "%-17s", "                 ");
    fd = open("/proc/net/arp", O_RDONLY);
    if (fd<0)    return 0;
    res = read(fd, buf, sizeof(buf)-1);
    close(fd);
    if (res <=0) return 0;
    buf[res] =0;

    pos = strstr(buf, ipaddr);
    while ( pos )
    {
        ptr = strtok(pos, " \t\r\n");
        if( strcmp(ptr, ipaddr) ) {
            pos = strstr(ptr+1, ipaddr);
            continue;
        }
        ptr = strtok(NULL, " \t");
        ptr = strtok(NULL, " \t");
        ptr = strtok(NULL, " \t");
        strncpy(mac, ptr, 17);
        return 1;
    }

    return 0;	
}	

#define INVALID_SOCKET	(-1)
char lan_subnet[20];
char lan_ip[20];

int get_lan_subnet (void)
{
    FILE *fp;
    char line[255];
    char *p;

    memset (lan_subnet, 0, sizeof (lan_subnet));
    memset (lan_ip, 0, sizeof (lan_ip));
    fp = fopen ("/var/tmp/hosts", "r");
    if (!fp) {
        kprintf ("cant' find hosts\n");
        strcpy (lan_subnet, "192.168.1.255");
        strcpy (lan_ip, "192.168.1.1");
        return -1;
    }
    memset (line, 0, sizeof (line));
    while (fgets (line, sizeof (line)-1, fp)) {
        if (strstr (line, "127.0.0.1")) {
            continue;
        }
        p = strtok (line, " \t");
        if (p) {
            strncpy (lan_ip, line, sizeof (lan_ip)-1);
            p = strrchr (line, '.');
            if (p) {
                p++;
                p[0] = '2';
                p[1] = '5';
                p[2] = '5';
                p[3] = 0;
                strncpy (lan_subnet, line, sizeof (lan_subnet) -1);
                break;
            }
        }
    } 
    fclose (fp);
    kprintf ("---got lan_ip %s,%d\n", lan_ip, strlen (lan_ip));
    if (strlen (lan_ip) == 0) {
        strcpy (lan_subnet, "192.168.1.255");
        strcpy (lan_ip, "192.168.1.1");
    }
    return 0;
}

int sock_138, sock_137;
struct sockaddr_in     addr_server_138, addr_server_137;

#define max(a,b) ((a)>(b))?(a):(b)
void *arp_creator (void)
{
	int             result;
	fd_set          fdset;								//file descripter set
	struct timeval  tval;								//timeout structure
	int             size;								//packet size
	char            response[20];
    int maxfdp1;
	int opt, num_try;
    struct sockaddr_in cli_addr;
    char ip[20];
    char mac[20];
    char cmd[100];
#ifndef STANDALONE
    sleep (30); //wait for /var/tmp/hosts be ready
    get_lan_subnet ();
#endif 
    num_try = 0;
    sock_138 = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock_138 == INVALID_SOCKET) {
        kprintf ("can't create socket for utg\n");
        return -1;
    }

    sock_137 = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock_137 == INVALID_SOCKET) {
        kprintf ("can   't create socket for 137\n");
        close (sock_138);
        return -1;
    }
    //socket for smb server
    opt = 1;
    memset (&addr_server_138, 0, sizeof (addr_server_138));
    addr_server_138.sin_family = AF_INET;
    addr_server_138.sin_addr.s_addr = inet_addr (lan_subnet);
    addr_server_138.sin_port = htons(138);

    if (bind (sock_138, (struct sockaddr *) &addr_server_138, sizeof (addr_server_138)) < 0 ) 
    {
        kprintf ("-----bind 138 error\n");
        close (sock_138);
        close (sock_137);
        return (-1);
    }
    //socket for smb server
    opt = 1;
    memset (&addr_server_137, 0, sizeof (addr_server_137));
    addr_server_137.sin_family = AF_INET;
    addr_server_137.sin_addr.s_addr = inet_addr (lan_subnet);
    addr_server_137.sin_port = htons(137);
    if (bind (sock_137, (struct sockaddr *) &addr_server_137, sizeof (addr_server_137)) < 0 ) 
    {
        kprintf ("-----bind 137 error\n");
        close (sock_138);
        close (sock_137);
        return (-1);
    }

    while (1) { 
        FD_ZERO(&fdset);						//reset the file descripter set
        maxfdp1 = 0;
    	FD_SET(sock_138,    &fdset);						//add the socket to the set
    	FD_SET(sock_137,    &fdset);						//add the socket to the set
    
    	tval.tv_sec  = 120;
    	tval.tv_usec = 0;
        maxfdp1 = max (sock_138, sock_137) + 1;
    	result = select (maxfdp1, &fdset, NULL, NULL, &tval);
        if (result < 0)
    	{
            kprintf ("----select error in arp creator\n");
    		break;
    	}
        if (result == 0) { //timeout
            //kprintf ("I'm still working %d\n", num_try ++);
            sleep (10); //in case something wrong , don't running without a rest
            continue;
        }
        
        if (FD_ISSET (sock_138, &fdset)) {
            int len = sizeof (cli_addr);

            //read the response packet
        	memset (response, 0, sizeof(response));
        	size = recvfrom(sock_138, (char *)response, sizeof (response)-1, 0, (struct sockaddr *)&cli_addr, &len);
            if (size > 0)
            {
                memset (ip, 0, sizeof (ip));
                strcpy (ip, inet_ntoa (cli_addr.sin_addr));
                //kprintf ("138--from %s, %u\n", inet_ntoa (cli_addr.sin_addr), time (NULL));
                if (!read_mac (ip, mac)) {
                    memset (cmd, 0, sizeof (cmd));
                    snprintf (cmd, sizeof (cmd) -1, "arping -q -c 1 -I br0 %s&", ip);
                    system (cmd);
                    kprintf ("send arp to %s\n", ip);
                }
            }
        }
        if (FD_ISSET (sock_137, &fdset)) {
            int len = sizeof (cli_addr);

            //read the response packet
        	memset (response, 0, sizeof(response));
        	size = recvfrom(sock_137, (char *)response, sizeof (response)-1, 0, (struct sockaddr *)&cli_addr, &len);
            if (size > 0)
            {
                memset (ip, 0, sizeof (ip));
                strcpy (ip, inet_ntoa (cli_addr.sin_addr));
                //kprintf ("137--from %s, %u\n", inet_ntoa (cli_addr.sin_addr), time (NULL));
                if (!read_mac (ip, mac)) {
                    memset (cmd, 0, sizeof (cmd));
                    snprintf (cmd, sizeof (cmd) -1, "arping -q -c 1 -I br0 %s&", ip);
                    system (cmd);
                    kprintf ("send arp -> %s\n", ip);
                }
            }
        }
        sleep (5); //if too many packets, take a breathe
    }
    close (sock_138);
    close (sock_137);
	return 0;
}

#ifdef STANDALONE
int main(int argc, char **argv)
{
    kprintf ("----arp creator\n");
    strcpy (lan_subnet, "192.168.1.255");
	arp_creator ();
}
#endif //STANDALONE
