/* dhcpd.h */
#ifndef _DHCPD_H
#define _DHCPD_H

#include <netinet/ip.h>
#include <netinet/udp.h>
#ifdef CONFIG_ACTIONTEC_TR111
#include <stdarg.h>
#include "packet.h"
#endif

#include "leases.h"

/************************************/
/* Defaults _you_ may want to tweak */
/************************************/

/* the period of time the client is allowed to use that address */
#define LEASE_TIME              (60*60*24*10) /* 10 days of seconds */

/* where to find the DHCP server configuration file */
#define DHCPD_CONF_FILE         "/etc/udhcpd.conf"

/*****************************************************************/
/* Do not modify below here unless you know what you are doing!! */
/*****************************************************************/

/* DHCP protocol -- see RFC 2131 */
#define SERVER_PORT		67
#define CLIENT_PORT		68

#define DHCP_MAGIC		0x63825363

/* DHCP option codes (partial list) */
#define DHCP_PADDING		0x00
#define DHCP_SUBNET		0x01
#define DHCP_TIME_OFFSET	0x02
#define DHCP_ROUTER		0x03
#define DHCP_TIME_SERVER	0x04
#define DHCP_NAME_SERVER	0x05
#define DHCP_DNS_SERVER		0x06
#define DHCP_LOG_SERVER		0x07
#define DHCP_COOKIE_SERVER	0x08
#define DHCP_LPR_SERVER		0x09
#define DHCP_HOST_NAME		0x0c
#define DHCP_BOOT_SIZE		0x0d
#define DHCP_DOMAIN_NAME	0x0f
#define DHCP_SWAP_SERVER	0x10
#define DHCP_ROOT_PATH		0x11
#define DHCP_IP_TTL		0x17
#define DHCP_MTU		0x1a
#define DHCP_BROADCAST		0x1c
#define DHCP_NTP_SERVER		0x2a
#define DHCP_WINS_SERVER	0x2c
#define DHCP_REQUESTED_IP	0x32
#define DHCP_LEASE_TIME		0x33
#define DHCP_OPTION_OVER	0x34
#define DHCP_MESSAGE_TYPE	0x35
#define DHCP_SERVER_ID		0x36
#define DHCP_PARAM_REQ		0x37
#define DHCP_MESSAGE		0x38
#define DHCP_MAX_SIZE		0x39
#define DHCP_T1			0x3a
#define DHCP_T2			0x3b
#define DHCP_VENDOR		0x3c
#define DHCP_CLIENT_ID		0x3d

/* compatibility for TR-111 */
#ifdef CONFIG_ACTIONTEC_TR111
#define DHCP_VENDOR_SPECIFIC_INFO 0x7d
#endif
/*  --- end --- 21/03/2007 */

#define DHCP_END		0xFF


#define BOOTREQUEST		1
#define BOOTREPLY		2

#define ETH_10MB		1
#define ETH_10MB_LEN		6

#define DHCPDISCOVER		1
#define DHCPOFFER		2
#define DHCPREQUEST		3
#define DHCPDECLINE		4
#define DHCPACK			5
#define DHCPNAK			6
#define DHCPRELEASE		7
#define DHCPINFORM		8

#define BROADCAST_FLAG		0x8000

#define OPTION_FIELD		0
#define FILE_FIELD		1
#define SNAME_FIELD		2

/* miscellaneous defines */
#define TRUE			1
#define FALSE			0
#define MAC_BCAST_ADDR		(unsigned char *) "\xff\xff\xff\xff\xff\xff"
#define OPT_CODE 0
#define OPT_LEN 1
#define OPT_DATA 2

/* miscellaneous defines */
/* compatibility for TR-111 */
#ifdef CONFIG_ACTIONTEC_TR111
#define DSLFORUM_ENTERPRISE_NO "3561"
#define VENDOR_INFORMATION_HEAD_LENGTH 6
#define ALL_SUB_OPTION_HEAD_LENGTH 6
#define ENTERPRISE_NO_LENGTH 4
#define OPTION_CODE_LENGTH 1
#define OPTION_DATA_LENGTH 1
#define SUB_OPTION_CODE_LENGTH 1
#define SUB_OPTION_DATA_LENGTH 1
#define DEVICE_MANUFACTURER_OUI 0x01
#define DEVICE_SERIAL_NUMBER 0x02
#define DEVICE_PRODUCT_CLASS 0x03
#define GATEWAY_MANUFACTURER_OUI 0x04
#define GATEWAY_SERIAL_NUMBER 0x05
#define GATEWAY_PRODUCT_CLASS 0x06

#define ACTION_ADD_VENDOR_INFO 0
#define ACTION_REMOVE_VENDOR_INFO 1
#endif
/* ---- end --- 21/03/2007 */

struct option_set {
	unsigned char *data;
	struct option_set *next;
};

//For static IP lease
struct static_lease {
	uint8_t *mac;
	uint32_t *ip;
	struct static_lease *next;
};

struct server_config_t {
	u_int32_t server;		/* Our IP, in network order */
	u_int32_t start;		/* Start address of leases, network order */
	u_int32_t end;			/* End of leases, network order */
	struct option_set *options;	/* List of DHCP options loaded from the config file */
	char *interface;		/* The name of the interface to use */
	int ifindex;			/* Index number of the interface to use */
	unsigned char arp[6];		/* Our arp address */
	unsigned long lease;		/* lease time in seconds (host order) */
	unsigned long max_leases; 	/* maximum number of leases (including reserved address) */
	char remaining; 		/* should the lease file be interpreted as lease time remaining, or
			 		 * as the time the lease expires */
	unsigned long auto_time; 	/* how long should udhcpd wait before writing a config file.
					 * if this is zero, it will only write one on SIGUSR1 */
	unsigned long decline_time; 	/* how long an address is reserved if a client returns a
				    	 * decline message */
	unsigned long conflict_time; 	/* how long an arp conflict offender is leased for */
	unsigned long offer_time; 	/* how long an offered address is reserved */
	unsigned long min_lease; 	/* minimum lease a client can request*/
	char *lease_file;
	char *pidfile;
	char *notify_file;		/* What to run whenever leases are written */
	u_int32_t siaddr;		/* next server bootp option */
	char *sname;			/* bootp server name */
	char *boot_file;		/* bootp boot file option */
	int  active;
	unsigned long inflease_time; 	/* how long is infinite leasetime */
  struct dhcpOfferedAddr *leases;  
    //For static IP lease
	struct static_lease *static_leases; /*List of ip/mac pairs to assign static leases */   
    struct static_lease *reserved_WanIp; /*List of wan ip in the unnumber mode*/
    //For actiontec extend
    char *actiontec_vip_enable;
    char *actiontec_vip_gw;
    char *actiontec_vip_netmask;
};	

extern struct server_config_t server_config[];
extern struct dhcpOfferedAddr *leases;
extern int no_of_ifaces;

#ifdef CONFIG_ACTIONTEC_TR111
void deal_vendor_data(struct dhcpMessage *packet, u_int32_t yiaddr, int code, int action);
void insert_into_file(u_int32_t yiaddr, unsigned char *oui, unsigned char *sno, unsigned char *pclass);
void delete_from_file(u_int32_t yiaddr, unsigned char *oui, unsigned char *sno, unsigned char *pclass);
int find_record_in_file(char *filename, char *fmt, ...);
#endif
#endif
